#! /usr/bin/env python

# 
# Copyright  2000-2001, GMD, Sankt Augustin
# -- German National Research Center for Information Technology 
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import sys
import os
import re
import getopt
import pprint
import string

from stat import ST_SIZE
from dvbobjects.utils import *
from dvbobjects.utils.SpecFile import *
from dvbobjects.DSMCC import BIOP
from dvbobjects.DVB.DataCarousel import TransactionId

######################################################################
# Constants

TIMEOUT            = 0xFFFFFFFF
BLOCK_SIZE         = 4066
TABLE_ID           = 0x3C

######################################################################
# Tunables

MAX_MODULE_SIZE    = 1024 * 64          # containing > 1 BIOP Message
OPT_MODULE_SIZE    = MAX_MODULE_SIZE    # Optimal ??? cf. MHP

assert OPT_MODULE_SIZE <= MAX_MODULE_SIZE

######################################################################
# Counters

MODULE_ID          = 0
KEY_SERIAL_NUMBER  = 0x00000000

######################################################################
# Globals

INPUT_DIR  = None
OUTPUT_DIR = None
theObjectCarousel = None

######################################################################
# Transaction IDs

DSI_TransactionId = TransactionId(
    version = 0,
    identification = 0,
    )

DII_TransactionId = TransactionId(
    version = 0,
    identification = 2,
    )

######################################################################
class Module:

    def __init__(self):
        global MODULE_ID
        self.module_id = MODULE_ID = MODULE_ID + 1

        fn = "%s/%04d.mod" % (OUTPUT_DIR, self.module_id)
        self.name = fn                  # needed for debug

        if DEBUG: print "NEW MODULE: %s" % fn

        self.__file = open(fn, "wb")
        self.size = 0

    def hasRoom(self, requestedSize):
        if self.size == 0:
            # Anything goes...
            return 1
        elif self.size + requestedSize <= OPT_MODULE_SIZE:
            # Module isn't empty, but...
            return 1
        else:
            # Start new module...
            return 0

    def write(self, bytes):
        msgSize = len(bytes)

        if DEBUG: print "ADD (mod %d, size %d+%d=%d)" % (
            self.module_id, self.size,
            msgSize, self.size+msgSize)

        self.__file.write(bytes)
        self.size = self.size + msgSize

######################################################################
class ObjectCarousel:

    # Just for fun...
    __singletons = []

    def __init__(self):
        assert not self.__singletons, "Singleton..."
        self.__singletons.append(self)

        self.__SPECdii = open("%s/DII.spec" % OUTPUT_DIR, "w")

        # Carousel Specification
        self.__initSpec()

        # currently open modules indexed by type
        self.__ModByType = {
            'dir': self.__nextModule(),
            'fil': self.__nextModule(),
	    'str': self.__nextModule(),
	    'ste': self.__nextModule(),
            }

        # Table of Contents, for debugging
        self.TOC = []

    def __initSpec(self):
        self.__spec = SuperGroupSpec(
            transactionId = DSI_TransactionId,
            srg_ior       = "%s/SRG_IOR" % OUTPUT_DIR,
            )

        self.__spec.addGroup(
            transactionId = DII_TransactionId,
            downloadId    = int(DOWNLOAD_ID),
            assocTag      = int(ASSOC_TAG, 16),
            blockSize     = BLOCK_SIZE,
            )

    def genSpec(self, OUTPUT_DIR):
        self.__spec.write(OUTPUT_DIR)
        
    def addDirectory(self, node):
        self.__addNode(node, "dir")

    def addFile(self, node):
        self.__addNode(node, "fil")

    def addStreamEvent(self, node):
        self.__addNode(node, "ste")

    def __addNode(self, node, type):
        msg = node.message()
        msgBytes = msg.pack()
        msgSize = len(msgBytes)

        if self.__ModByType[type].hasRoom(msgSize):
            self.__ModByType[type].write(msgBytes)
        else:
            self.__ModByType[type] = self.__nextModule()
            self.__ModByType[type].write(msgBytes)

        node.bind(
            carouselId = int(CAROUSEL_ID),
            moduleId   = self.__ModByType[type].module_id,
            )

        self.TOC.append((
            self.__ModByType[type].module_id,
            msg.objectKey,
            os.path.basename(msg.PATH),
            msg.objectKind,
            msgSize,
            ))

    def __nextModule(self):
        mod = Module()
        self.__SPECdii.write("%s 0x%02X 0x%04X 0x%02X\n" % (
            mod.name,
            TABLE_ID, mod.module_id, int(MODULE_VERSION, 16)))
        self.__spec.addModule(
            tableId       = TABLE_ID,
            moduleId      = mod.module_id,
            moduleVersion = int(MODULE_VERSION, 16),
            )
        return mod

######################################################################
class FSNode(DVBobject):

    """A File System Node destined for an ObjectCarousel.
    Superclass for FSDirectory, FSSteam and FSFile.
    """

    def __init__(self):
        global KEY_SERIAL_NUMBER
        self.KEY = KEY_SERIAL_NUMBER = KEY_SERIAL_NUMBER + 1
        
    def IOR(self, carouselId, moduleId, key):
        iop = BIOP.IOP.IOR(
            PATH = self.PATH,           # for debugging
            type_id = self.MessageClass.objectKind,

            # Object Location attributes
            carouselId = carouselId,
            moduleId   = moduleId,
            objectKey  = key,

            # ConnBinder / Delivery Tap attributes
            assocTag      = int(ASSOC_TAG, 16),
            transactionId = DII_TransactionId,
            timeout       = TIMEOUT,
            )

        return iop

    def _checkBinding(self):
        try:
            raise "Already Bound", self._binding
        except AttributeError:
            pass
        
######################################################################
class FSFile(FSNode):

    """A File in a File System destined for an Object Carousel.
    """

    MessageClass = BIOP.FileMessage
    BindingClass = BIOP.ObjectFileBinding
   
    def __init__(self, path):
        FSNode.__init__(self)
        self.PATH = path
        self.contentSize = os.stat(path)[ST_SIZE]

    def bind(self, carouselId, moduleId):
        self._checkBinding()
        fn = os.path.basename(self.PATH)
        self._binding = self.BindingClass(
            nameId = fn + "\x00",
            IOR = self.IOR(carouselId, moduleId, self.KEY),
            contentSize=self.contentSize,
            )

    def binding(self):
        return self._binding

    def message(self):
        msg = self.MessageClass(
            PATH = self.PATH,
            objectKey   = self.KEY,
            contentSize = self.contentSize,
            )
        return msg

    def shipMessage(self):
        theObjectCarousel.addFile(self)
            
######################################################################
class FSStreamEvent(FSNode):

    """A Directory in a File System destined to genereate a StreamEvent Object for Object Carousel.
    """

    MessageClass = BIOP.StreamEventMessage
    BindingClass = BIOP.ObjectStreamEventBinding
    
    def __init__(self, path):
        FSNode.__init__(self)
	self.PATH = path

    def bind(self, carouselId, moduleId):
        self._checkBinding()
        fn = os.path.basename(self.PATH)
        self._binding = self.BindingClass(
            nameId = fn + "\x00",
            IOR = self.IOR(carouselId, moduleId, self.KEY),
    	    )

    def binding(self):
        return self._binding

    def message(self):
        msg = self.MessageClass(
            PATH = self.PATH,
            objectKey   = self.KEY,
            )
        return msg

    def shipMessage(self):
        theObjectCarousel.addStreamEvent(self)
            

######################################################################
class FSDir(FSNode):

    """A Directory in a File System destined for an Object Carousel.
    """

    MessageClass = BIOP.DirectoryMessage
    BindingClass = BIOP.ContextBinding

    def __init__(self, path):
        FSNode.__init__(self)
        self.bindings = []
        self.PATH  = path
        assert(len(path) > 0)

    def bind(self, carouselId, moduleId):
        self._checkBinding()
        fn = os.path.basename(self.PATH)
#        print "Fehler: ", self.PATH, fn
        self._binding = self.BindingClass(
            nameId = fn + "\x00",
            IOR = self.IOR(carouselId, moduleId, self.KEY),
            )

    def binding(self):
        return self._binding

    def message(self):
        msg = self.MessageClass(
            PATH = self.PATH,
            objectKey = self.KEY,
            bindings = self.bindings,
            )
        return msg 
        
    def visit(self):
        #REJECT_EXT = ['.pyc', '.o', '.so']
        # ... oder so:
        #REJECT_EXT = []
        
        #REJECT_FN = ['x', 'tmp']
        # ... oder so:
        #REJECT_FN = []
        
	EVENT_EXT = ['.event']
	
        assert os.path.isdir(self.PATH), self.PATH
        try:
            ls = os.listdir(self.PATH)
        except:
            print self.PATH
            raise
        
        ls.sort()

        for fn in ls:
            path = os.path.join(self.PATH, fn)

            #if fn in REJECT_FN:
            #    continue
            #elif os.path.splitext(fn)[1] in REJECT_EXT:
            #    continue
            ###elif os.path.islink(path):
            ###   continue          
            if os.path.isfile(path):
		obj = FSFile(path)
    		obj.shipMessage()
        	if DEBUG:
            	    print obj.message()
            	    print
            elif os.path.isdir(path):
		if os.path.splitext(fn)[1] in EVENT_EXT:
		    # Stream event
		    obj = FSStreamEvent(path)
            	    obj.shipMessage()
            	    if DEBUG:
                	print obj.message()
                	print
		else:
            	    # Depth first visit		
		    obj = FSDir(path)
            	    obj.visit()
            	    if DEBUG:
                	print obj.message()
                	print			    
            else:
                continue               

            self.bindings.append(obj.binding())

        # THIS directory (i.e. self) is complete, so...
        self.shipMessage()
        if DEBUG:
            print self.message()
            print
            
    def shipMessage(self):
        theObjectCarousel.addDirectory(self)
            
######################################################################
class FSRoot(FSDir):

    """A Directory in a File System.
    Destined as Service Gateway for an Object Carousel.
    """

    MessageClass = BIOP.ServiceGatewayMessage

######################################################################
OPTIONS = "hd"
LONG_OPTIONS = [
    "help",
    "debug",
    ]

def Usage(return_code = 1):
    print ("Usage: %s"
           " <InputDirectory>"
           " <OutputModulesDirectory>"
	   " download_id"
	   " carousel_id"
	   " association_tag"
	   " version" ) % (
        sys.argv[0])
    sys.exit(return_code)

def CheckArgs():
    global INPUT_DIR, OUTPUT_DIR, DEBUG, CAROUSEL_ID, DOWNLOAD_ID, ASSOC_TAG, MODULE_VERSION 
    
    try:
        opts, args = getopt.getopt(
            sys.argv[1:], OPTIONS, LONG_OPTIONS)
    except getopt.error:
        Usage()

    DEBUG = 0
    for opt_name, opt_val in opts:
        if opt_name in ['-d', '--debug']:
            DEBUG = 1
        elif opt_name in ['-h', '--help']:
            Usage(0)

    if len(args) <> 6:
        Usage()

    INPUT_DIR, OUTPUT_DIR, CAROUSEL_ID, DOWNLOAD_ID, ASSOC_TAG, MODULE_VERSION = args

    DII_TransactionId.version = int(MODULE_VERSION, 16)
    DSI_TransactionId.version = DII_TransactionId.version

######################################################################
def Main():

    global theObjectCarousel
    theObjectCarousel = ObjectCarousel()

    root = FSRoot(INPUT_DIR)
    root.visit()
    srg_ior = root.binding().IOR
    open("%s/SRG_IOR" % OUTPUT_DIR, "wb").write(srg_ior.pack())
    theObjectCarousel.genSpec(OUTPUT_DIR)

    if DEBUG:
        print ; print srg_ior
        pprint.pprint(theObjectCarousel.TOC)

######################################################################
if __name__ == '__main__':
    CheckArgs()
#    print OUTPUT_DIR
    Main()
